;;;; Clipboard functions for Corman Lisp - Version 1.0
;;;;
;;;; Copyright (C) 2002 Christopher Double. All Rights Reserved.
;;;; 
;;;; License
;;;; =======
;;;; This software is provided 'as-is', without any express or implied
;;;; warranty. In no event will the author be held liable for any damages
;;;; arising from the use of this software.
;;;;
;;;; Permission is granted to anyone to use this software for any purpose,
;;;; including commercial applications, and to alter it and redistribute
;;;; it freely, subject to the following restrictions:
;;;;
;;;; 1. The origin of this software must not be misrepresented; you must
;;;;    not claim that you wrote the original software. If you use this
;;;;    software in a product, an acknowledgment in the product documentation
;;;;    would be appreciated but is not required.
;;;;
;;;; 2. Altered source versions must be plainly marked as such, and must
;;;;    not be misrepresented as being the original software.
;;;;
;;;; 3. This notice may not be removed or altered from any source 
;;;;    distribution.
;;;;
;;;; Notes
;;;; =====
;;;; Inspired by some Forth code I came across for evaluating and
;;;; manipulating items on the clipboard. The Forth code was from:
;;;;
;;;; http://home.earthlink.net/~neilbawd/lnfilt.html
;;;; http://home.earthlink.net/~neilbawd/lnfiltex.html
;;;;
;;;; This package allows reading and writing text to the clipboard,
;;;; as well as evaluating forms held in the clipboard.
;;;;
;;;; More recent versions of this software may be available at:
;;;;   http://www.double.co.nz/cl
;;;;
;;;; Comments, suggestions and bug reports to the author, 
;;;; Christopher Double, at: chris.double@double.co.nz
;;;;
;;;; 05/04/2002 - 1.0 
;;;;              Initial release.
(defpackage "CLIPBOARD"
	(:use 
		"COMMON-LISP"
		"WIN32"
		"CT")
	(:export 
		"STRING-FROM-CLIPBOARD"
		"STRING-TO-CLIPBOARD"
		"EVALUATE-CLIPBOARD"
		"EC"
		"DO-CLIPBOARD-LINES"))

(in-package :clipboard)

#! (:library "KERNEL32" :ignore "WINBASEAPI" :export t :pascal "WINAPI")
WINBASEAPI
void*
WINAPI
GlobalAlloc(
    UINT uFlags,
    DWORD dwBytes
    );

WINBASEAPI
void*
WINAPI
GlobalFree(
    HANDLE hMem
    );

WINBASEAPI
LPVOID
WINAPI
GlobalLock(
    HANDLE hMem
    );

WINBASEAPI
BOOL
WINAPI
GlobalUnlock(
    void* hMem
    );

#define GMEM_MOVEABLE 2
!#

#! (:library "USER32" :ignore "WINUSERAPI" :export t :pascal "WINAPI")
WINUSERAPI
BOOL
WINAPI
OpenClipboard(
    HWND hWndNewOwner);

WINUSERAPI
BOOL
WINAPI
CloseClipboard();


WINUSERAPI
BOOL
WINAPI
IsClipboardFormatAvailable(
    UINT format);

WINUSERAPI
HANDLE
WINAPI
    GetClipboardData(
    UINT uFormat);

WINUSERAPI
HANDLE
WINAPI
SetClipboardData(
    UINT uFormat,
    HANDLE hMem);

WINUSERAPI
BOOL
WINAPI
EmptyClipboard();

#define CF_TEXT             1
!#

(defmacro with-clipboard (&body body)
	"Execute the body between calls of OpenClipboard and CloseClipboard"
	`(if (OpenClipboard null)
		(unwind-protect
			(progn
				,@body)
			(unless (CloseClipboard)
				(error "CloseClipboard failed")))
		(error "OpenClipboard failed")))

(defun string-from-clipboard ()
	(with-clipboard
		(let ((data (getclipboarddata CF_TEXT)))
			(when (cpointer-null data)
				(error "GetClipboardData returned NULL"))
			(c-string-to-lisp-string data))))

(defmacro with-global-lock ((ptr hglobal) &body body)
	"Execute the body between calls of GlobalLock and GlobalUnlock"
	`(let ((,ptr (GlobalLock ,hglobal)))
		(if (cpointer-null ,ptr)
			(error "GlobalLock failed")
			(unwind-protect
				(progn ,@body)
				(unless (GlobalUnlock ,ptr)
					(error "GlobalUnlock failed"))))))

(defun string-to-clipboard (str)
	(with-clipboard
		(let ((hglobal (GlobalAlloc GMEM_MOVEABLE (+ (length str) 1))))
			(when (cpointer-null hglobal)
				(error "GlobalAlloc failed"))
			(with-global-lock (ptr hglobal)
				(lisp-string-to-c-string str ptr))
			(when (cpointer-null (SetClipboardData CF_TEXT hglobal))
				(error "SetClipboardData failed")))))
	
(defun evaluate-clipboard ()
	(let* ((str (string-from-clipboard))
			(len (length str)))
		(loop 
			for (form index) = (multiple-value-list (read-from-string str))
			then (multiple-value-list (read-from-string str nil nil :start index))
			until (>= index len)
			do (eval form)
			finally (eval form))))
   
(defun ec ()
	"Shortened form of evaluate-clipboard for easy typing"
	(evaluate-clipboard))

(defun do-clipboard-lines (func)
	"For each line in the clipboard, execute (FUNC line)"
	(let ((stream (make-string-input-stream (string-from-clipboard))))
		(loop
			as line = (read-line stream nil)
			while line
			do (funcall func line))))

;; Examples
(defun comment-filter ()
	"Place a comment block around the lines in the clipboard. 
Store the result back in the clipboard"
	(string-to-clipboard
		(with-output-to-string (stream)
			(write-line "#|" stream)
			(do-clipboard-lines #'(lambda (line)
					(write-line line stream)))
			(write-line "|#" stream))))

(defun macroexpand-filter ()
	"Take the lisp code in the clipboard, macroexpand it, store the
result in the clipboard"
	(string-to-clipboard
   		(with-output-to-string (stream)
			(pprint 
				(macroexpand
			    	(read-from-string (string-from-clipboard))) stream))))
					 

(provide 'clipboard)

